
// Standardne pre MS win aplikacie
#include <windows.h>

// GLEW kniznica. pre spravu novsich rozsireni
#include "glew.h"
// GL Utility Toolkit (GLUT), v ramci toho sa includuje aj openGL hlavickovy subor
#include "glut.h"

#include <stdio.h>
#include <math.h>

#include "shaders.h"
#include "textures.h"
#include "geometry.h"
#include "fonts.h"
#include "meshes.h"
#include "pose.h"
#include "fbo.h"

// Globalne premenne
GLboolean g_is_anim = true;
GLboolean g_is_texture = false;
GLboolean g_is_linear_interpolation = false;
float g_current_time = 0.0f;

// Globalne premenne pre prezeranie sceny
GLdouble g_vzd, g_old_vzd, g_fi, g_xi, g_old_fi, g_old_xi;
GLint g_left_mouse_down_x, g_left_mouse_down_y;
GLint g_right_mouse_down_y;
GLboolean g_right_down = false, g_left_down = false;
GLuint g_window_height, g_window_width;

// id textur
GLuint g_tex_sun, g_tex_earth, g_tex_lava, g_tex_phobos, g_tex_skybox;

// id shaderov
GLuint g_shaders_skybox, g_shaders_perpixel_phong, g_shaders_envmap;
GLuint g_shaders_mrt_skybox, g_shaders_mrt_phong;

// modely
std::vector<ObjectPose> g_duck_animation_keys;
std::vector<float> g_duck_animation_knots;
Scene* g_duck_model = NULL;
Scene* g_car_model = NULL;

// framebuffer objekty
FBO g_info_fbo;
FBO g_mirror_fbo;
#define MIRROR_TEX_SIZE 512
std::vector<aiVector3D> g_mirror_points;

void InitAnimations()
{
    g_duck_animation_knots.push_back(0);
    g_duck_animation_keys.push_back(CreatePose(4, 0, 0, 90, 0, 90, 0.01f, 0.01f, 0.01f));
    g_duck_animation_knots.push_back(3);
    g_duck_animation_keys.push_back(CreatePose(4, 4, 0, 90, 0, 135, 0.01f, 0.01f, 0.01f));
    g_duck_animation_knots.push_back(6);
    g_duck_animation_keys.push_back(CreatePose(-4, 4, 0, 90, 0, 220, 0.01f, 0.01f, 0.01f));
    g_duck_animation_knots.push_back(9);
    g_duck_animation_keys.push_back(CreatePose(-4, -4, 0, 90, 0, 310, 0.01f, 0.01f, 0.01f));
    g_duck_animation_knots.push_back(12);
    g_duck_animation_keys.push_back(CreatePose(4, -4, 0, 90, 0, 40, 0.01f, 0.01f, 0.01f));
    g_duck_animation_knots.push_back(15);
    g_duck_animation_keys.push_back(CreatePose(4, 0, 0, 90, 0, 90, 0.01f, 0.01f, 0.01f));
}

void initShaders()
{
    // nacitanie, kompilacia a zlinkovanie shaderov
    g_shaders_skybox = loadProgram("shaders/skybox.vert", "shaders/skybox.frag");
    SetShaderUniform1i(g_shaders_skybox, "cube_map", 0);

    g_shaders_perpixel_phong = loadProgram("shaders/perpixel_phong.vert", "shaders/perpixel_phong.frag");
    SetShaderUniform1i(g_shaders_perpixel_phong, "color_texture", 0);

    g_shaders_envmap = loadProgram("shaders/perpixel_envmap.vert", "shaders/perpixel_envmap.frag");
    SetShaderUniform1i(g_shaders_envmap, "color_texture", 0);
    SetShaderUniform1i(g_shaders_envmap, "cubemap_texture", 1);

    g_shaders_mrt_skybox = loadProgram("shaders/mrt_skybox.vert", "shaders/mrt_skybox.frag");
    SetShaderUniform1i(g_shaders_mrt_skybox, "cube_map", 0);

    g_shaders_mrt_phong = loadProgram("shaders/mrt_phong.vert", "shaders/mrt_phong.frag");
    SetShaderUniform1i(g_shaders_mrt_phong, "color_texture", 0);

    glActiveTexture(GL_TEXTURE0);
}

// inicializacia OpenGL textur
void initTextures()
{
    // ************* prva textura, zem *************
    g_tex_earth = LoadTexture("textures/earth.jpg");

    // ************* druha textura, slnko *************
    g_tex_sun = LoadTexture("textures/sun.jpg");

    // ************* tretia textura, jupiter *************
    g_tex_lava = LoadTexture("textures/lava.png");

    // ************* stvrta textura, phobos *************
    g_tex_phobos = LoadTexture("textures/phobos.jpg");

    // ************* stvrta textura, phobos *************
    g_tex_skybox = LoadCubeTexture("skybox/snow_positive_x.jpg", "skybox/snow_negative_x.jpg",
                                 "skybox/snow_positive_y.jpg", "skybox/snow_negative_y.jpg",
                                 "skybox/snow_positive_z.jpg", "skybox/snow_negative_z.jpg");
}

// Open GL Inicializacia
bool initGL(void)
{
    // vypis info o o systeme OpenGL
    printf("OpenGL version: %s\n", glGetString(GL_VERSION));
    printf("OpenGL vendor: %s\n", glGetString(GL_VENDOR));
    printf("OpenGL renderer: %s\n", glGetString(GL_RENDERER));

    GLint maxColorAttachments;
    glGetIntegerv(GL_MAX_COLOR_ATTACHMENTS_EXT, &maxColorAttachments);
    printf("Max FBO color attachments: %i\n", maxColorAttachments);

    // nastav cierne pozadie
    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    
    // nastav z-buffer testovanie
    glClearDepth(1.0f);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);

    // nastav stencil buffer
    glClearStencil(0);

    // nastav zmiesavaciu funkciu
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    // nastav vhodne mapovanie textur
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

    // initializuj kameru
    g_fi = 0.7f; g_xi = 0.7f; g_vzd = 10.0f;

    // initializuje 1 svetlo
    GLfloat light_ambient[] = { 0.2f, 0.2f, 0.2f, 1.0f };
    GLfloat light_diffuse[] = { 1.0f, 1.0f, 1.0f, 1.0f };
    GLfloat light_specular[] = { 0.6f, 0.6f, 0.6f, 1.0f };
    glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
    glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
    glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, 1.0);
    glEnable(GL_LIGHT0);

    // sposob zmiesavania farby textury a povodnej farby, pri pouzivani shaderov sa toto nepouziva
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

    // inicializuj FBO
    g_info_fbo = CreateFBO(MIRROR_TEX_SIZE, MIRROR_TEX_SIZE, 2, true);
    g_mirror_fbo = CreateFBO(MIRROR_TEX_SIZE, MIRROR_TEX_SIZE, 1, false);
    // initializuj vrcholy zrkadla
    g_mirror_points.push_back(aiVector3D(-3, -3, 0));
    g_mirror_points.push_back(aiVector3D(3, -3, 0));
    g_mirror_points.push_back(aiVector3D(3, 3, 0));
    g_mirror_points.push_back(aiVector3D(-3, 3, 0));

	return true;
}

// vyrenderuj celu scenu, modelview je v tomto momente uz nastavena a obsahuje view maticu kamery
void renderWorldScene(bool mrt_pass)
{
    // vykresli suradnicove osi svetoveho systemu roznymi farbami, bez shaderov
    glUseProgram(0);
    glDisable(GL_TEXTURE_2D);
    glLineWidth(1);
    glBegin( GL_LINES );
    glColor3f(1.0f, 0.0f, 0.0f );glVertex3f( 0.0f, 0.0f,0.0f );glVertex3f( 100.0f, 0.0f, 0.0f );
    glColor3f(0.0f, 1.0f, 0.0f );glVertex3f( 0.0f, 0.0f,0.0f );glVertex3f( 0.0f, 100.0f, 0.0f );
    glColor3f(0.0f, 0.0f, 1.0f );glVertex3f( 0.0f, 0.0f,0.0f );glVertex3f( 0.0f, 0.0f, 100.0f );
    glEnd();

    // OBJEKTY POUZIVAJUCE SKYBOX SHADER
    // skybox je dany vo svetovych suradniciach, akurat jeho stred je stale v bode, kde je kamera
    GLuint current_shader = 0;
    if (!mrt_pass)
        current_shader = g_shaders_skybox;
    else
        current_shader = g_shaders_mrt_skybox;
    glUseProgram(current_shader);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_CUBE_MAP, g_tex_skybox);
    glPushMatrix();
    glScalef(100, 100, 100);
    renderCubeVBO();
    glPopMatrix();
    glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
    glUseProgram(0);

    // TERAZ IDU OBJEKTY POUZIVAJUCE PER-PIXEL PHONG SHADER
    if (!mrt_pass)
        current_shader = g_shaders_perpixel_phong;
    else
        current_shader = g_shaders_mrt_phong;
    glUseProgram(current_shader);
    if (g_is_texture)
        SetShaderUniform1i(current_shader, "texturing_enabled", 1);
    else
        SetShaderUniform1i(current_shader, "texturing_enabled", 0);

    // vykresli slnko
    SetShaderUniform1i(current_shader, "lighting_enabled", 0);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, g_tex_sun);
    glPushMatrix();
    glRotatef(100*g_current_time, 0.0f, 0.0f, 1.0f);
    glTranslatef(8.0f, 0.0f, 0.0f);
    // nastav poziciu svetla do stredu slnka
    GLfloat sun_position[] = { 0.0f, 0.0f, 0.0f, 1.0f };
    glLightfv(GL_LIGHT0, GL_POSITION, sun_position);
    glScalef(0.3f, 0.3f, 0.3f);
    glColor3f(1.0f, 1.0f, 1.0f );
    renderSphereVBO();
    glPopMatrix();

    // vykresli kacicu
    SetShaderUniform1i(current_shader, "lighting_enabled", 1);
    glPushMatrix();
    // nastav pozu kacice
    ObjectPose duck_pose;
    if (g_is_linear_interpolation)
        duck_pose = IntepolatePoseLinear(g_duck_animation_knots, g_duck_animation_keys, g_current_time, false);
    else
        duck_pose = IntepolatePoseCubic(g_duck_animation_knots, g_duck_animation_keys, g_current_time, false);
    // z pozy vytvor a nahraj modelovaciu maticu
    aiMatrix4x4 duck_model_matrix = GetMatrix(duck_pose);
    MultiplyMatrixOpenGL(duck_model_matrix);
    // vykresli 
    RenderScene(g_duck_model, current_shader);
    glPopMatrix();

    // vykresli klucove body cesty pre kacku ako dalsie kacky
    for (int i = 0; i < (int)g_duck_animation_keys.size(); i++)
    {
        glPushMatrix();
        aiMatrix4x4 duck_model_matrix = GetMatrix(g_duck_animation_keys[i]);
        MultiplyMatrixOpenGL(duck_model_matrix);
        RenderScene(g_duck_model, current_shader);
        glPopMatrix();
    }

    // vykresli auto
    glPushMatrix();
    ObjectPose car_pose = CreatePose(0, 0, 0.5f, 90, 0, 0, 0.2f, 0.2f, 0.2f);
    MultiplyMatrixOpenGL(GetMatrix(car_pose)); 
    RenderScene(g_car_model, current_shader);
    glPopMatrix();

    // vykresli 3D text, bez shaderov
    glUseProgram(NULL);
    glEnable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);
    glColor3f(1.0f, 0.0f, 0.0f);
    render3DTextWGL(5, 0, 2, 1, "Project GRC - FBO ...");
    glDisable(GL_LIGHTING);
}

// Procedura obsluhujuca vykreslovanie
void render(void)   
{
    // Vypocitaj poziciu kamery v svetovych suradniciach
    GLfloat camera_pos[4];
    camera_pos[0] = g_vzd * cos(g_fi) * cos(g_xi);
	camera_pos[1] = g_vzd * sin(g_fi) * cos(g_xi);
	camera_pos[2] = g_vzd * sin(g_xi);
    camera_pos[3] = 1;
    
    // KRESLIME SCENU DO TEXTURY SPOZA ZRKADLA
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, g_mirror_fbo.id);
    // Vymaz (vypln) FBO bufre definovanymi hodnotami
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    glPushAttrib(GL_VIEWPORT_BIT | GL_COLOR_BUFFER_BIT);
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glViewport(0, 0, MIRROR_TEX_SIZE, MIRROR_TEX_SIZE);
    // nastav matice pre pohlad spoza zrkadla
    SetMirrorMatrices(aiVector3D(camera_pos[0], camera_pos[1], camera_pos[2]), g_mirror_points);
    // zapamataj si matice
    GLdouble mirror_view_matrix[16];
    glGetDoublev(GL_MODELVIEW_MATRIX, mirror_view_matrix);
    GLdouble mirror_projection_matrix[16];
    glGetDoublev(GL_PROJECTION_MATRIX, mirror_projection_matrix);
    // vykresli scenu
    renderWorldScene(false);
    // vrat sa k povodnym hodnotam transformacii
    glPopAttrib();
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();

    // KRESLIME SCENU DO OKNA
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
    // Vymaz (vypln) oknove bufre definovanymi hodnotami
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);	
    // Nastav view maticu pre pohlad na svetovy suradnicovy system
    glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
    ObjectPose camera_pose = CreateLookAtPose(camera_pos[0], camera_pos[1], camera_pos[2], 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f);
    MultiplyMatrixOpenGL(GetMatrix(camera_pose).Inverse());
    // vykresli scenu
    renderWorldScene(false);
    // vykresli zrkadlo
    RenderMirror(g_mirror_points, mirror_view_matrix, mirror_projection_matrix, g_mirror_fbo.color_textures[0]);

    // VYRENDERUJEME TEXTURY S NORMALAMI, DIFUZNOU ZLOZKOU A DEPTH BUFROM
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, g_info_fbo.id);
    // Vymaz (vypln) oknove bufre definovanymi hodnotami
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    glPushAttrib(GL_VIEWPORT_BIT | GL_COLOR_BUFFER_BIT);
    glViewport(0, 0, MIRROR_TEX_SIZE, MIRROR_TEX_SIZE);
    GLuint attachments[4] = { GL_COLOR_ATTACHMENT0, GL_COLOR_ATTACHMENT1, GL_COLOR_ATTACHMENT2, GL_COLOR_ATTACHMENT3 };
    glDrawBuffers(2, attachments);
    renderWorldScene(true);
    glPopAttrib();

    // OPAT KRESLIME SCENU DO OKNA
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
    // vykresli 2D text, bez shaderov
    glColor3f(1.0f, 1.0f, 0.0f);
    render2DTextFT(10, g_window_height - 20, "Projekt GRC ...");
    render2DTextFT(10, g_window_height - 45, "F1 - zapni/vypni animaciu");
    render2DTextFT(10, g_window_height - 70, "F2 - zapni/vypni textury");
    render2DTextFT(10, g_window_height - 95, "F3 - zapni/vypni linearnu interpolaciu");

    // vykresli info textury do okna
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glOrtho(0, g_window_width, 0, g_window_height, -1, 1);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glBindTexture(GL_TEXTURE_2D, g_info_fbo.color_textures[0]);
    glEnable(GL_TEXTURE_2D);
    glBegin(GL_QUADS);
    glTexCoord2f(0, 0);glVertex3f(0,0,0);
    glTexCoord2f(1, 0);glVertex3f(0.3f*g_window_width,0,0);
    glTexCoord2f(1, 1);glVertex3f(0.3f*g_window_width,0.3f*g_window_height,0);
    glTexCoord2f(0, 1);glVertex3f(0,0.3f*g_window_height,0);
    glEnd();
    glBindTexture(GL_TEXTURE_2D, g_info_fbo.color_textures[1]);
    glBegin(GL_QUADS);
    glTexCoord2f(0, 0);glVertex3f(0.3f*g_window_width,0,0);
    glTexCoord2f(1, 0);glVertex3f(0.6f*g_window_width,0,0);
    glTexCoord2f(1, 1);glVertex3f(0.6f*g_window_width,0.3f*g_window_height,0);
    glTexCoord2f(0, 1);glVertex3f(0.3f*g_window_width,0.3f*g_window_height,0);
    glEnd();
    glBindTexture(GL_TEXTURE_2D, g_info_fbo.depth_texture);
    glBegin(GL_QUADS);
    glTexCoord2f(0, 0);glVertex3f(0.6f*g_window_width,0,0);
    glTexCoord2f(1, 0);glVertex3f(0.9f*g_window_width,0,0);
    glTexCoord2f(1, 1);glVertex3f(0.9f*g_window_width,0.3f*g_window_height,0);
    glTexCoord2f(0, 1);glVertex3f(0.6f*g_window_width,0.3f*g_window_height,0);
    glEnd();
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);

    // Zamen buffre a zobraz vykreslene objekty
    glutSwapBuffers();
}

// Procedura obsluhujuca zmenu rozmerov okna
void reshape(int w, int h)
{
    g_window_width = w;
    g_window_height = h;

    // nastvenie casti okna pre vykreslenie, tentoraz kreslime do celeho okna
	glViewport(0, 0, w, h);
    // Budeme menit projekcnu maticu (transformaciu)
	glMatrixMode(GL_PROJECTION);                        
	glLoadIdentity();                                  

	if (h == 0) 
        h = 1;
    // Chceme perspektivnu projekciu
    float ratio = (float)w/(float)h;
	gluPerspective(60, ratio, 0.1, 300.0);
    //glOrtho(-10.0f*ratio, 10.0f*ratio, -10.0f, 10.0f, 1.0, 5000.0); 
}

// Procedura obsluhujuca stlacenie klavesy
void keyboard(unsigned char key, int x, int y)
{
	switch (key) 
    {
        // Ak Esc bolo stlacene
		case 27:
            DeleteAllVBO();
            if (g_duck_model)
                delete g_duck_model;
            if (g_car_model)
                delete g_car_model;
            glDeleteFramebuffersEXT(1, &g_mirror_fbo.id);
            // Ukoncenie programu
			exit(0);
		    break;
		default:        
		    break;
	}
}

// Procedura obsluhujuca stlacenie specialnej klavesy
void special_keys(int a_keys, int x, int y)
{
	switch (a_keys) 
    {
        // Ak F1 bolo stlacene
		case GLUT_KEY_F1:
            g_is_anim = !g_is_anim;
            glutPostRedisplay();
		    break;
        // Ak F2 bolo stlacene
        case GLUT_KEY_F2:
            g_is_texture = !g_is_texture;
            glutPostRedisplay();
		    break;
        // Ak F3 bolo stlacene
        case GLUT_KEY_F3:
            g_is_linear_interpolation = !g_is_linear_interpolation;
            glutPostRedisplay();
		    break;
		default:
		    break;
	}
}

// Procedura obsluhujuca stlacenie tlacitka mysi
void mouse_klik(int button, int state, int x, int y )
{
	switch (button) 
    {
		case GLUT_LEFT_BUTTON:
            if (state == GLUT_DOWN)
            {
                g_left_down = true;
			    g_left_mouse_down_x = x;
			    g_left_mouse_down_y = y;
			    g_old_fi = g_fi;
			    g_old_xi = g_xi;
                break;
            }
            else if (state == GLUT_UP)
            {
                g_left_down = false;
            }
		    break;
        case GLUT_RIGHT_BUTTON:
            if (state == GLUT_DOWN)
            {
                g_right_down = true;
			    g_right_mouse_down_y = y;
			    g_old_vzd = g_vzd;
                break;
            }
            else if (state == GLUT_UP)
            {
                g_right_down = false;
            }
		    break;
		default:
		    break;
	}

    glutPostRedisplay();
}

// Procedura obsluhujuca pohyb mysi
void mouse_move(int x, int y)
{
    if (g_left_down == GL_TRUE)
    {
        g_fi = g_old_fi + (g_left_mouse_down_x - x)/200.0f;
        g_xi = g_old_xi + (y - g_left_mouse_down_y)/200.0f;
        glutPostRedisplay();
    }

    if (g_right_down == GL_TRUE)
    {
        g_vzd = g_old_vzd + (g_right_mouse_down_y - y)/10.0f;
        if (g_vzd > 80) g_vzd = 80;
        glutPostRedisplay();
    } 

}

void timer(int id)
{
    if (g_is_anim)
    {
        g_current_time += 0.050f;
    }
    else
    {
        //g_current_time = 0.0f;
    }

    // jednoduche opakovanie animacie po 17 sekundach
    if (g_current_time > 17)
        g_current_time = 0.0f;

    glutPostRedisplay();
    glutTimerFunc(50, timer, 0);
}

// Zakladna funkcia spustena na zaciatku
int main(int argc, char** argv)
{
    // GLUT Inicializacia
	glutInit(&argc, argv);                              
    // Display Mod (Rgb a Double Buffering)
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH | GLUT_STENCIL);
    //glutInitDisplayString("stencil>=8 rgb>=8 double depth>=24");
    // Nastavime pociatocnu poziciu a velkost okna
    glutInitWindowPosition(50, 50);
    glutInitWindowSize(800, 600);                   
    // Vyrobime okno s danym menom
    glutCreateWindow("--== GRC projekt ==--" );  

    // Zinicializujeme GLEW
    glewInit();

    // Nastavime zakladne vlastnosti OpenGL
	initGL();
    initSphereVBO(32, 32);
    initTorusVBO(0.6f, 2.0f, 32, 32);
    initTextures();
    initShaders();
    init2DFontFT("fonts/optien.ttf");
    InitAnimations();

    // nacitaj modely
    g_duck_model = LoadScene("./models/Duck/duck.dae");
    g_car_model = LoadScene("./models/F40/F40.dae");

    // Zaregistrujeme callback funkcie
	glutDisplayFunc(render);                            
	glutReshapeFunc(reshape);                        
	glutKeyboardFunc(keyboard);                      
	glutSpecialFunc(special_keys);
    glutMouseFunc(mouse_klik);
    glutMotionFunc(mouse_move);
    glutTimerFunc(0, timer, 0);

    // Spustenie hlavneho okruhu zachytavania sprav
	glutMainLoop();

	return 0;
}
